"use strict";

exports.__esModule = true;
exports.useState = useState;
exports.useContext = useContext;
exports.useEffect = useEffect;
exports.useLayoutEffect = useLayoutEffect;
exports.useImperativeHandle = useImperativeHandle;
exports.useRef = useRef;
exports.useCallback = useCallback;
exports.useMemo = useMemo;
exports.useReducer = useReducer;

var _host = _interopRequireDefault(require("./vdom/host"));

var _scheduler = require("./vdom/scheduler");

var _shallowEqual = require("./vdom/shallowEqual");

var _types = require("./types");

var _error = require("./error");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function getCurrentInstance() {
  return _host.default.owner && _host.default.owner._instance;
}

function getCurrentRenderingInstance() {
  var currentInstance = getCurrentInstance();

  if (currentInstance) {
    return currentInstance;
  } else {
    if (process.env.NODE_ENV === 'production') {
      (0, _error.invokeMinifiedError)(1);
    } else {
      throw new Error('Hooks can only be called inside a component.');
    }
  }
}

function areInputsEqual(inputs, prevInputs) {
  if (prevInputs === null || inputs.length !== prevInputs.length) {
    return false;
  }

  for (var i = 0; i < inputs.length; i++) {
    if ((0, _shallowEqual.is)(inputs[i], prevInputs[i])) {
      continue;
    }

    return false;
  }

  return true;
}

function useState(initialState) {
  var currentInstance = getCurrentRenderingInstance();
  var hookID = currentInstance.getHookID();
  var hooks = currentInstance.getHooks();

  if (!hooks[hookID]) {
    // If the initial state is the result of an expensive computation,
    // you may provide a function instead for lazy initial state.
    if ((0, _types.isFunction)(initialState)) {
      initialState = initialState();
    }

    var setState = function setState(newState) {
      // Flush all effects first before update state
      if (!_host.default.isUpdating) {
        (0, _scheduler.flushEffect)();
      }

      var hook = hooks[hookID];
      var eagerState = hook[2]; // function updater

      if ((0, _types.isFunction)(newState)) {
        newState = newState(eagerState);
      }

      if (!(0, _shallowEqual.is)(newState, eagerState)) {
        // Current instance is in render update phase.
        // After this one render finish, will containue run.
        hook[2] = newState;

        if (getCurrentInstance() === currentInstance) {
          // Marked as is scheduled that could finish hooks.
          currentInstance.isScheduled = true;
        } else {
          currentInstance.update();
        }
      }
    };

    hooks[hookID] = [initialState, setState, initialState];
  }

  var hook = hooks[hookID];

  if (!(0, _shallowEqual.is)(hook[0], hook[2])) {
    hook[0] = hook[2];
    currentInstance.shouldUpdate = true;
  }

  return hook;
}

function useContext(context) {
  var currentInstance = getCurrentRenderingInstance();
  return currentInstance.readContext(context);
}

function useEffect(effect, inputs) {
  useEffectImpl(effect, inputs, true);
}

function useLayoutEffect(effect, inputs) {
  useEffectImpl(effect, inputs);
}

function useEffectImpl(effect, inputs, defered) {
  var currentInstance = getCurrentRenderingInstance();
  var hookID = currentInstance.getHookID();
  var hooks = currentInstance.getHooks();
  inputs = inputs === undefined ? null : inputs;

  if (!hooks[hookID]) {
    var create = function create(immediately) {
      if (!immediately && defered) return (0, _scheduler.scheduleEffect)(function () {
        return create(true);
      });
      var current = create.current;

      if (current) {
        destory.current = current();
        create.current = null;
      }
    };

    var destory = function destory(immediately) {
      if (!immediately && defered) return (0, _scheduler.scheduleEffect)(function () {
        return destory(true);
      });
      var current = destory.current;

      if (current) {
        current();
        destory.current = null;
      }
    };

    create.current = effect;
    hooks[hookID] = {
      create: create,
      destory: destory,
      prevInputs: inputs,
      inputs: inputs
    };
    currentInstance.didMount.push(create);
    currentInstance.willUnmount.push(destory);
    currentInstance.didUpdate.push(function () {
      var _hooks$hookID = hooks[hookID],
          prevInputs = _hooks$hookID.prevInputs,
          inputs = _hooks$hookID.inputs,
          create = _hooks$hookID.create;

      if (inputs == null || !areInputsEqual(inputs, prevInputs)) {
        destory();
        create();
      }
    });
  } else {
    var hook = hooks[hookID];
    var _create = hook.create,
        prevInputs = hook.inputs;
    hook.inputs = inputs;
    hook.prevInputs = prevInputs;
    _create.current = effect;
  }
}

function useImperativeHandle(ref, create, inputs) {
  var nextInputs = inputs != null ? inputs.concat([ref]) : null;
  useLayoutEffect(function () {
    if ((0, _types.isFunction)(ref)) {
      ref(create());
      return function () {
        return ref(null);
      };
    } else if (ref != null) {
      ref.current = create();
      return function () {
        ref.current = null;
      };
    }
  }, nextInputs);
}

function useRef(initialValue) {
  var currentInstance = getCurrentRenderingInstance();
  var hookID = currentInstance.getHookID();
  var hooks = currentInstance.getHooks();

  if (!hooks[hookID]) {
    hooks[hookID] = {
      current: initialValue
    };
  }

  return hooks[hookID];
}

function useCallback(callback, inputs) {
  return useMemo(function () {
    return callback;
  }, inputs);
}

function useMemo(create, inputs) {
  var currentInstance = getCurrentRenderingInstance();
  var hookID = currentInstance.getHookID();
  var hooks = currentInstance.getHooks();
  inputs = inputs === undefined ? null : inputs;

  if (!hooks[hookID]) {
    hooks[hookID] = [create(), inputs];
  } else {
    var prevInputs = hooks[hookID][1];

    if (inputs === null || !areInputsEqual(inputs, prevInputs)) {
      hooks[hookID] = [create(), inputs];
    }
  }

  return hooks[hookID][0];
}

function useReducer(reducer, initialArg, init) {
  var currentInstance = getCurrentRenderingInstance();
  var hookID = currentInstance.getHookID();
  var hooks = currentInstance.getHooks();

  if (!hooks[hookID]) {
    var initialState = init !== undefined ? init(initialArg) : initialArg;

    var dispatch = function dispatch(action) {
      // Flush all effects first before update state
      if (!_host.default.isUpdating) {
        (0, _scheduler.flushEffect)();
      }

      var hook = hooks[hookID]; // Reducer will update in the next render, before that we add all
      // actions to the queue

      var queue = hook[2];

      if (getCurrentInstance() === currentInstance) {
        queue.actions.push(action);
        currentInstance.isScheduled = true;
      } else {
        var currentState = queue.eagerState;
        var eagerReducer = queue.eagerReducer;
        var eagerState = eagerReducer(currentState, action);

        if ((0, _shallowEqual.is)(eagerState, currentState)) {
          return;
        }

        queue.eagerState = eagerState;
        queue.actions.push(action);
        currentInstance.update();
      }
    };

    return hooks[hookID] = [initialState, dispatch, {
      actions: [],
      eagerReducer: reducer,
      eagerState: initialState
    }];
  }

  var hook = hooks[hookID];
  var queue = hook[2];
  var next = hook[0];

  if (currentInstance._reRenders > 0) {
    for (var i = 0; i < queue.actions.length; i++) {
      next = reducer(next, queue.actions[i]);
    }
  } else {
    next = queue.eagerState;
  }

  if (!(0, _shallowEqual.is)(next, hook[0])) {
    hook[0] = next;
    currentInstance.shouldUpdate = true;
  }

  queue.eagerReducer = reducer;
  queue.eagerState = next;
  queue.actions.length = 0;
  return hooks[hookID];
}